"""
Environment
====================================
Pythonic representation of the NetWitness Environment
"""

import faker
import paramiko
import click
import json
import getpass
import logging
import os.path
from ipaddress import IPv4Address

fake = faker.Faker()
fake.add_provider(faker.providers.internet)


class Host():
    """Object representation of a NetWitness host

    Attributes
    ----------
    ipv4 : str
        Host IPv4 address.
    hostname : str
        Host hostname.
    minion_id : str
        Host salt minion ID
    head : Head
        Head Unit controlling this host
    """

    __slots__ = [
        '_head',
        '_ipv4',
        '_hostname',
        '_minion_id',
        '_services',
        '_manual', 
        '_env'
    ]

    def __init__(self, ipv4=str, minion_id=str, head=None, _manual=False, _env=None, hostname=str, services=None):
        self._head = head
        self._ipv4 = ipv4
        self._hostname = hostname
        self._minion_id = minion_id
        self._services = services
        self._manual = _manual
        self._env = _env

    def exec(self, command, saltArgs=None):
        if saltArgs is None:
            saltArgs = ''
        stdin, stdout, stderr = self.head._exec(f"salt \"{self.minion_id}\" cmd.run \"{command}\" {saltArgs}")
        return stdin, stdout, stderr

    @property
    def hostname(self):
        return self._hostname

    @property
    def minion_id(self):
        return self._minion_id

    @property
    def services(self):
        return self._services

    @property
    def ipv4(self):
        return self._ipv4

    @property
    def head(self):
        return self._head


class Head(Host):
    """Object representation of a NetWitness head unit

    Attributes
    ----------
    _client : paramiko.SSHClient
        Backend SSH connection to the head unit
    """

    def __init__(self, ipv4, _env):
        ipv4 = str(ipv4)
        self._client = paramiko.SSHClient()
        self._client.load_system_host_keys()
        self._client.set_missing_host_key_policy(paramiko.WarningPolicy)
        try:
            self._client.connect(ipv4, username='root', banner_timeout=200)
        except paramiko.PasswordRequiredException:
            self._client.connect(ipv4, username='root',
                                 password=getpass.getpass('SSH Key Password: '), banner_timeout=200)
        except paramiko.AuthenticationException:
            self._client.connect(ipv4, username='root',
                                 password=getpass.getpass('Root Password: '), banner_timeout=200)
        except paramiko.SSHException:
            self._client.connect(ipv4, username='root',
                                 password=getpass.getpass('Root Password: '), banner_timeout=200)

        self.__sftp = None
        super().__init__(ipv4=ipv4, head=self, _env=_env)

    def _exec(self, command, saltArgs=None):
        stdin, stdout, stderr = self._client.exec_command(command)
        return stdin, stdout, stderr

    def exec_many(self, targets, command, filter='-L', saltArgs=None):
        pass

    def send_file(self, f):
        logging.debug(f)
        f.seek(0)
        self._sftp.putfo(f, f'/tmp/{os.path.basename(f.name)}')

    @property
    def _sftp(self):
        if self.__sftp is None:
            self.__sftp = self._client.open_sftp()
        return self.__sftp


class Environment():
    """Object representation of the NetWitness environment
    """

    def __init__(self, head=None):
        if head is None:
            head = click.prompt('Head unit IP', type=IPv4Address)
        self.head = Head(head, self)
        self._deploy_admin = None
        self.hosts = self.getAllHosts()

    @property
    def deploy_admin(self):
        if self._deploy_admin is None:
            _, stdout, _ = self.head._exec(
                'security-cli-client --get-config-prop --prop-hierarchy nw.security-client --prop-name platform.deployment.password --quiet')
            self._deploy_admin = stdout.read()
            if isinstance(self._deploy_admin, bytes):
                self._deploy_admin = self._deploy_admin.decode('utf-8')
            self._deploy_admin = self._deploy_admin.replace('\n', '')
        return self._deploy_admin

    def getAllHosts(self):
        hosts = [self.head]
        _, stdout, _ = self.head._exec(f'mongoexport --type=json -d orchestration-server -c host -u deploy_admin -p {self.deploy_admin} --authenticationDatabase=admin -f _id,installedServices,displayName,ipv4,hostname --quiet')
        #print(f'mongoexport --type=json -d orchestration-server -c host -u deploy_admin -p {self.deploy_admin} --authenticationDatabase=admin -f _id,installedServices,displayName,ipv4 --quiet')
        for line in stdout:
            l = json.loads(line)
            if l.get('displayName') is None:
                l['displayName'] = l['hostname']
            if l.get('ipv4') is None:
                l['ipv4'] = l['hostname']
            if l['ipv4'] == self.head._ipv4:
                self.head._minion_id = l['_id']
                self.head._hostname = l['displayName']
                self.head._services = [service.lower() for service in l['installedServices']]
            else:
                hosts.append(
                    Host(minion_id=l['_id'], head=self.head, ipv4=l['ipv4'], _env=self, hostname=l['displayName'],
                         services=l['installedServices']))
        return hosts
