#!/usr/bin/python


'''
Author: Mohamed Osama
Email: mohamed.osama@rsa.com
Creation Date: 3/4/2020
Last Modification Date: 7/6/2020
Purpose: Perform Health Check in All Netwitness Environment
Features:
    - Perform Full Health Check* if you chose option 0
    * Full Health Check would include all features pointed out in below menu
    - You can separately check each Item as listed in below menu
1. Environment Info
2. Hosts' Specifications; ex: Highest Disk Usage & Free Memory Left
3. DNS IP
4. SELinux status
5. Appliance Serial Number
6. NTP Status and current Date
7. Routing Table
8. RabbitMQ Users and Permissions
9. Existence of Core Files
10. Failed Services
11. Current Error logs in /var/log/messages
'''


import click
import nwtoolkit.color as color
import pathlib
from nwtoolkit.reporting import Report
import re


def report(data, output_format):
    template_path = f'{pathlib.Path(__file__).parent.absolute()}\\'
    report = Report(template_path, 'health_check_template.htm')
    if output_format == 'report_file':
        report.to_file('health_check.htm', data=data, overwrite=True)
    else:
        report.to_stdout(data=data)

@click.command(help='Health Check on Admin Server and all Netwitness hosts')
@click.pass_obj
def health_check(env):
    menu(env)


def full(env):
    data = {}
    filter = re.compile("\w{8}-\w{4}-\w{4}-\w{4}-\w{12}")

    print(color.green("Listing all Netwitness Hosts IPs and their Salt Minion IDs..."))
    for host in env.hosts:
        print(color.blue(f'{host.minion_id} {host.hostname} {host.ipv4}'))
        data[host.minion_id] = {}
        data[host.minion_id]['hostname'] = host.hostname
        data[host.minion_id]['IPV4'] = host.ipv4


    print(color.green("Highest Disk usage on all hosts"))
    Usage = None
    for host in env.hosts:
        Usage = []
        _, output, _ = host.exec("df -h | awk '{ print \$5,\$6 }' | sort -n -r | sed '8d'")
        print(color.blue(f'{host.minion_id} '), end='')
        for line in output:
            if filter.match(line):
                continue
            Usage.append(line)
            print(color.blue(f'{line}'), end='')
            data[host.minion_id]['disk_usage'] = ''.join(Usage)


    print(color.green("Free Memory left on all hosts (Gigabytes)"))
    for host in env.hosts:
        _, output, _ = host.exec("free -g | awk '{ print \$4 }'  | head -n 2 | sed '1d'")
        for line in output:
            if filter.match(line):
                continue
            print(color.blue(f'{host.minion_id} {line}'), end='')
            data[host.minion_id]['Free_Memory_Left'] = line

    print(color.green("Service currently consuming most of host memory"))
    MEM = None
    for host in env.hosts:
        MEM = []
        _, output, _ = host.exec("ps -eo pid,user,cmd,%mem --sort=-%mem | head -2")
        print(color.blue(f'{host.minion_id} '), end='')
        for line in output:
            if filter.match(line):
                continue
            MEM.append(line)
            print(color.blue(f'{line}'), end='')
            data[host.minion_id]['Top_Memory_Service'] = ''.join(MEM)

    print(color.green("Host's DNS IP"))
    for host in env.hosts:
        _, output, _ = host.exec("cat /etc/resolv.conf | grep -i nameserver | sed -e 's/\<nameserver\>//g'")
        for line in output:
            if filter.match(line):
                continue
            print(color.blue(f'{host.minion_id} {line}'), end='')
            data[host.minion_id]['DNS'] = line


    print(color.green("SELinux status"))
    for host in env.hosts:
        _, output, _ = host.exec("getenforce")
        for line in output:
            if filter.match(line):
                continue
            print(color.blue(f'{host.minion_id} {line}'), end='')
            data[host.minion_id]['selinux'] = line


    print(color.green("Host's Serial Number"))
    for host in env.hosts:
        _, output, _ = host.exec("dmidecode -s system-serial-number")
        for line in output:
            if filter.match(line):
                continue
            print(color.blue(f'{host.minion_id} {line}'), end='')
            data[host.minion_id]['SN'] = line


    print(color.green("NTP Status"))
    NTP = None
    for host in env.hosts:
        NTP = []
        _, output, _ = host.exec("ntpstat")
        print(color.blue(f'{host.minion_id} '), end='')
        for line in output:
            if filter.match(line):
                continue
            NTP.append(line)
            print(color.blue(f'{line}'), end='')
            data[host.minion_id]['NTP'] = ''.join(NTP)


    print(color.green("Current Date & Time"))
    for host in env.hosts:
        _, output, _ = host.exec("date")
        for line in output:
            if filter.match(line):
                continue
            print(color.blue(f'{host.minion_id} {line}'), end='')
            data[host.minion_id]['Date'] = line


    print(color.green("Routing Table"))
    RT = None
    for host in env.hosts:
        RT = []
        _, output, _ = host.exec("route -n")
        print(color.blue(f'{host.minion_id} '), end='')
        for line in output:
            if filter.match(line):
                continue
            RT.append(line)
            print(color.blue(f'{line}'), end='')
            data[host.minion_id]['RT'] = ''.join(RT)


    print(color.green("Rabbitmq users on /rsa/system vhost"))
    RU = None
    for host in env.hosts:
        RU = []
        _, output, _ = host.exec("rabbitmqctl list_users")
        print(color.blue(f'{host.minion_id} '), end='')
        for line in output:
            if filter.match(line):
                continue
            RU.append(line)
            print(color.blue(f'{line}'), end='')
            data[host.minion_id]['RU'] = ''.join(RU)


    print(color.green("Rabbitmq permissions on /rsa/system vhost"))
    RP = None
    for host in env.hosts:
        RP = []
        _, output, _ = host.exec("rabbitmqctl list_permissions -p /rsa/system")
        print(color.blue(f'{host.minion_id} '), end='')
        for line in output:
            if filter.match(line):
                continue
            RP.append(line)
            print(color.blue(f'{line}'), end='')
            data[host.minion_id]['RP'] = ''.join(RP)


    print(color.green("Existence of Core Files on all hosts"))
    Core = None
    for host in env.hosts:
        Core = []
        printed = False
        _, output, _ = host.exec("find /var/netwitness/ -iname \"core*.log\"")
        for line in output:
            if filter.match(line):
                continue
            Core.append(line)
            print(color.red(f'{host.minion_id} {line}'), end='')
            data[host.minion_id]['Existence_Core_Files'] = ''.join(Core)
            printed = True
        if not printed:
            print(color.blue(f'{host.minion_id} \tNo core files found'))
            data[host.minion_id]['Existence_Core_Files'] = 'No core files found\n'


    print(color.green("Failed services on all hosts"))
    Service = None
    for host in env.hosts:
        Service = []
        printed = False
        _, output, _ = host.exec("systemctl | grep -i failed")
        for line in output:
            if filter.match(line):
                continue
            Service.append(line)
            print(color.red(f'{host.minion_id} {line}'), end='')
            data[host.minion_id]['Failed_Services'] = ''.join(Service)
            printed = True
        if not printed:
            print(color.blue(f'{host.minion_id} \tNo Failed Services on this host'))
            data[host.minion_id]['Failed_Services'] = 'No Failed Services on this host\n'


    print(color.green("Error Logs in latest 20 lines of /var/log/messages"))
    LOGS = None
    for host in env.hosts:
        LOGS = []
        _, output, _ = host.exec("tail -20 /var/log/messages | grep -i error | grep -v salt-minion")
        print(color.blue(f'{host.minion_id} '), end='')
        for line in output:
            if filter.match(line):
                continue
            LOGS.append(line)
            print(color.blue(f'{line}'), end='')
            data[host.minion_id]['LOGS'] = ''.join(LOGS)


    report(data, 'report_file')

def info(env):
    data = {}
    filter = re.compile("\w{8}-\w{4}-\w{4}-\w{4}-\w{12}")
    print(color.green("Listing all Netwitness Hosts IPs and their Salt Minion IDs..."))
    for host in env.hosts:
        print(color.blue(f'{host.minion_id} {host.hostname} {host.ipv4}'))
        data[host.minion_id] = {}
        data[host.minion_id]['hostname'] = host.hostname
        data[host.minion_id]['IPV4'] = host.ipv4
    report(data, 'report_file')

def specs(env):
    data = {}
    filter = re.compile("\w{8}-\w{4}-\w{4}-\w{4}-\w{12}")

    print(color.green("Highest Disk usage on all hosts"))
    Usage = None
    for host in env.hosts:
        Usage = []
        _, output, _ = host.exec("df -h | awk '{ print \$5,\$6 }' | sort -n -r | sed '8d'")
        print(color.blue(f'{host.minion_id} '), end='')
        for line in output:
            if filter.match(line):
                continue
            Usage.append(line)
            print(color.blue(f'{line}'), end='')
            data[host.minion_id] = {}
            data[host.minion_id]['disk_usage'] = ''.join(Usage)
    report(data, 'report_file')



    print(color.green("Free Memory left on all hosts (Gigabytes)"))
    for host in env.hosts:
        _, output, _ = host.exec("free -g | awk '{ print \$4 }'  | head -n 2 | sed '1d'")
        for line in output:
            if filter.match(line):
                continue
            print(color.blue(f'{host.minion_id} {line}'), end='')
            data[host.minion_id]['Free_Memory_Left'] = line
    report(data, 'report_file')

    print(color.green("Service currently consuming most of host memory"))
    MEM = None
    for host in env.hosts:
        MEM = []
        _, output, _ = host.exec("ps -eo pid,user,cmd,%mem --sort=-%mem | head -2")
        print(color.blue(f'{host.minion_id} '), end='')
        for line in output:
            if filter.match(line):
                continue
            MEM.append(line)
            print(color.blue(f'{line}'), end='')
            data[host.minion_id]['Top_Memory_Service'] = ''.join(MEM)
    report(data, 'report_file')


def dns(env):
    data = {}
    filter = re.compile("\w{8}-\w{4}-\w{4}-\w{4}-\w{12}")
    print(color.green("Host's DNS IP"))
    for host in env.hosts:
        _, output, _ = host.exec("cat /etc/resolv.conf | grep -i nameserver | sed -e 's/\<nameserver\>//g'")
        for line in output:
            if filter.match(line):
                continue
            print(color.blue(f'{host.minion_id} {line}'), end='')
            data[host.minion_id] = {}
            data[host.minion_id]['DNS'] = line
    report(data, 'report_file')

def selinux(env):
    data = {}
    filter = re.compile("\w{8}-\w{4}-\w{4}-\w{4}-\w{12}")
    print(color.green("SELinux status"))
    for host in env.hosts:
        _, output, _ = host.exec("getenforce")
        for line in output:
            if filter.match(line):
                continue
            print(color.blue(f'{host.minion_id} {line}'), end='')
            data[host.minion_id] = {}
            data[host.minion_id]['selinux'] = line
    report(data, 'report_file')

def serial(env):
    data = {}
    filter = re.compile("\w{8}-\w{4}-\w{4}-\w{4}-\w{12}")
    print(color.green("Host's Serial Number"))
    for host in env.hosts:
        _, output, _ = host.exec("dmidecode -s system-serial-number")
        for line in output:
            if filter.match(line):
                continue
            print(color.blue(f'{host.minion_id} {line}'), end='')
            data[host.minion_id] = {}
            data[host.minion_id]['SN'] = line
    report(data, 'report_file')

def date(env):
    data = {}
    filter = re.compile("\w{8}-\w{4}-\w{4}-\w{4}-\w{12}")
    print(color.green("NTP Status"))
    NTP = None
    for host in env.hosts:
        NTP = []
        _, output, _ = host.exec("ntpstat")
        print(color.blue(f'{host.minion_id} '), end='')
        for line in output:
            if filter.match(line):
                continue
            NTP.append(line)
            print(color.blue(f'{line}'), end='')
            data[host.minion_id] = {}
            data[host.minion_id]['NTP'] = ''.join(NTP)


    print(color.green("Current Date & Time"))
    for host in env.hosts:
        _, output, _ = host.exec("date")
        for line in output:
            if filter.match(line):
                continue
            print(color.blue(f'{host.minion_id} {line}'), end='')
            data[host.minion_id]['Date'] = line
    report(data, 'report_file')

def routing(env):
    data = {}
    filter = re.compile("\w{8}-\w{4}-\w{4}-\w{4}-\w{12}")
    print(color.green("Routing Table"))
    RT = None
    for host in env.hosts:
        RT = []
        _, output, _ = host.exec("route -n")
        print(color.blue(f'{host.minion_id} '), end='')
        for line in output:
            if filter.match(line):
                continue
            RT.append(line)
            print(color.blue(f'{line}'), end='')
            data[host.minion_id] = {}
            data[host.minion_id]['RT'] = ''.join(RT)
    report(data, 'report_file')

def rabbit(env):
    data = {}
    filter = re.compile("\w{8}-\w{4}-\w{4}-\w{4}-\w{12}")
    print(color.green("Rabbitmq users on /rsa/system vhost"))
    RU = None
    for host in env.hosts:
        RU = []
        _, output, _ = host.exec("rabbitmqctl list_users")
        print(color.blue(f'{host.minion_id} '), end='')
        for line in output:
            if filter.match(line):
                continue
            RU.append(line)
            print(color.blue(f'{line}'), end='')
            data[host.minion_id] = {}
            data[host.minion_id]['RU'] = ''.join(RU)

    print(color.green("Rabbitmq permissions on /rsa/system vhost"))
    RP = None
    for host in env.hosts:
        RP = []
        _, output, _ = host.exec("rabbitmqctl list_permissions -p /rsa/system")
        print(color.blue(f'{host.minion_id} '), end='')
        for line in output:
            if filter.match(line):
                continue
            RP.append(line)
            print(color.blue(f'{line}'), end='')
            data[host.minion_id]['RP'] = ''.join(RP)
    report(data, 'report_file')

def core(env):
    data = {}
    filter = re.compile("\w{8}-\w{4}-\w{4}-\w{4}-\w{12}")
    print(color.green("Existence of Core Files on all hosts"))
    Core = None
    for host in env.hosts:
        Core = []
        printed = False
        _, output, _ = host.exec("find /var/netwitness/ -iname \"core*.log\"")
        for line in output:
            if filter.match(line):
                continue
            Core.append(line)
            print(color.red(f'{host.minion_id} {line}'), end='')
            data[host.minion_id] = {}
            data[host.minion_id]['Existence_Core_Files'] = ''.join(Core)
            printed = True
        if not printed:
            print(color.blue(f'{host.minion_id} \tNo core files found'))
            data[host.minion_id] = {}
            data[host.minion_id]['Existence_Core_Files'] = 'No core files found\n'
    report(data, 'report_file')

def services(env):
    data = {}
    filter = re.compile("\w{8}-\w{4}-\w{4}-\w{4}-\w{12}")
    print(color.green("Failed services on all hosts"))
    Service = None
    for host in env.hosts:
        Service = []
        printed = False
        _, output, _ = host.exec("systemctl | grep -i failed")
        for line in output:
            if filter.match(line):
                continue
            Service.append(line)
            print(color.red(f'{host.minion_id} {line}'), end='')
            data[host.minion_id] = {}
            data[host.minion_id]['Failed_Services'] = ''.join(Service)
            printed = True
        if not printed:
            print(color.blue(f'{host.minion_id} \tNo Failed Services on this host'))
            data[host.minion_id] = {}
            data[host.minion_id]['Failed_Services'] = 'No Failed Services on this host\n'
    report(data, 'report_file')

def Logs(env):
    data = {}
    filter = re.compile("\w{8}-\w{4}-\w{4}-\w{4}-\w{12}")
    print(color.green("Error Logs in latest 20 lines of /var/log/messages"))
    LOGS = None
    for host in env.hosts:
        LOGS = []
        _, output, _ = host.exec("tail -20 /var/log/messages | grep -i error | grep -v salt-minion")
        print(color.blue(f'{host.minion_id} '), end='')
        for line in output:
            if filter.match(line):
                continue
            LOGS.append(line)
            print(color.blue(f'{line}'), end='')
            data[host.minion_id] = {}
            data[host.minion_id]['LOGS'] = ''.join(LOGS)
    report(data, 'report_file')

def menu(env):
    print(color.green("Welcome to Health Check.."))
    print(color.blue("=========================="))
    print(color.yellow("0. Full Health Check"))
    print(color.yellow("1. Environment Info"))
    print(color.yellow("2. Hosts' Specifications; ex: Highest Disk Usage & Free Memory Left"))
    print(color.yellow("3. DNS IP"))
    print(color.yellow("4. SELinux status"))
    print(color.yellow("5. Appliance Serial Number"))
    print(color.yellow("6. NTP Status and current Date"))
    print(color.yellow("7. Routing Table"))
    print(color.yellow("8. RabbitMQ Users and Permissions"))
    print(color.yellow("9. Existence of Core Files"))
    print(color.yellow("10. Failed Services"))
    print(color.yellow("11. Current Error logs in /var/log/messages"))
    print(color.blue("==============================================================="))
    choice = input("Please enter the number corresponding to the needed Health Check: ")
    if choice == "0":
        full(env)

    if choice == "1":
        info(env)

    if choice == "2":
        specs(env)

    if choice == "3":
        dns(env)

    if choice == "4":
        selinux(env)

    if choice == "5":
        serial(env)

    if choice == "6":
        date(env)

    if choice == "7":
        routing(env)

    if choice == "8":
        rabbit(env)

    if choice == "9":
        core(env)

    if choice == "10":
        services(env)

    if choice == "11":
        Logs(env)