'''
Author: Cody Spooner
Date: 6/10/20
Purpose: Test firewall ports for new installs and existing environments.
Features:
    - 1:1 testing (single target, can be used with listener command opensock)
    - Read in text comma delimited text file (can be used with listener command opensock) 
        Ex: 
        --> 192.168.10.13,Decoder
        --> 192.168.10.31,Concentrator
    - Directionality testing (existing environments only)
        - Tests all appliances in an environment by appliance type and tests connection between two appliances, not from Head Unit
'''

import json
import os
import sys
import nwtoolkit.color as color
from nwtoolkit.constants.ports import Components, Services, Appliance_Sockets, Direction
import click
import pathlib
from nwtoolkit.reporting import Report
import nwtoolkit.transform as magical
import pandas as pd
from tabulate import tabulate
import timeit
import logging
logging.getLogger("scapy.runtime").setLevel(logging.ERROR)
from scapy.all import *
from socket import *
from collections import defaultdict

def menu(): # main menu
    disclaimer = color.yellow(f"\n**This plugin is intended for use with an environment that is running version 11.3 or 11.4**\n")
    click.echo(disclaimer)
    data = {'1': 'New install FW Scan (to be used with opensock command)', '2': 'Existing environment scan'}
    table_data = pd.DataFrame.from_dict(data, orient='index',columns=['Main Menu'])
    table = (tabulate(table_data, headers='keys', tablefmt='psql'))
    click.echo(color.yellow(table))

def sub_menu(): # 1:1 sub menu, appliance list
    app_types = Appliance_Sockets
    app_type_dict = []
    apptype_json = json.dumps(app_types)
    app_type_dict = json.loads(apptype_json)
    menu_pretty = {} # stores data for pretty menu
    num = 0
    for apptype in app_type_dict.keys():
        app_type_dict[apptype] = num
        num+=1
    for name, choice_num in app_type_dict.items():
        menu_pretty[choice_num] = name
    table_data = pd.DataFrame.from_dict(menu_pretty, orient='index', columns=['Appliance Type']) 
    table = (tabulate(table_data, headers='keys', tablefmt='psql'))
    click.echo(color.yellow(f'\n{table}'))
    return app_type_dict # returns dictionary of tuples Ex. adminserver:'0'
        
def scanner(env,salt_cmd,head_id,test_port):
    stdin, stdout, stderr = env.head.exec(salt_cmd)
    for line in stdout:
        successful = False
        if "Successfully" in line:
            useful_output = color.green(f'Port: {test_port} is reachable \u2714')
            successful = True          
        elif "Unable" in line:
            useful_output = color.red(f'Port: {test_port} is not reachable \u274C')
        else:
            continue
        click.echo(useful_output)
        return successful

def ip_scan(env,app_ipv4,test_port,failcheck): #scanning function for reading in IPs from list or 1:1
    alive = ''
    dead = ''
    caution = ''
    if failcheck == True: #check if results are accurate. SCAPY FAILS against listener
        s = socket(AF_INET, SOCK_STREAM)
        s.settimeout(3)
        conn = s.connect_ex((app_ipv4, int(test_port)))
        if(conn == 0) :
            alive = test_port
        else:
            dead = test_port
        s.close()
    else:
        conf.verb = 0 # Disable verbose in sr(), sr1() methods
        ans,unans=sr(IP(dst=app_ipv4)/TCP(dport=int(test_port),flags="S"),timeout=3)
        for snd,rcv in ans:
            if("SA" in rcv.summary()):
                alive = test_port
            elif("RA" in rcv.summary()):
                caution = test_port
        for s in unans:
            dead = test_port
    return alive, caution, dead

def direction_scan(env,targ_serv, source, dst_minionid, ports,service): #automated scanning function using salt minions, no list input
    head_id = env.head.minion_id
    info = color.blue(f"Testing connection on {targ_serv} ({dst_minionid}) from {service} ({source})")
    click.echo(info)
    report_data = []
    report_dict = {}
    for test_port in sorted(ports):
        salt_cmd=(f"salt '{source}' network.connect {dst_minionid} {test_port}")
        result = scanner(env,salt_cmd, head_id,test_port)
        if(result == True):
            data = (f'{test_port} \u2714')
            report_data.append(data)
        else:
            report_data.append(data)
            data = (f'{test_port} \u274C')
    report_list = list(set(report_data))
    return report_list

def getType(env,targ_serv): #filters hosts by appliance type. Ex. returns minionids of all Brokers.
    dst_ids = []
    for host in env.hosts:
        service = host.services[0]
        if targ_serv == service:
            #print(f"{targ_serv} --- {service}") #DEBUG <-- uncomment to debug services not testing
            dst_ids.append(host.minion_id)
            #print(dst_ids) #DEBUG <-- uncomment to debug services not testing
        elif targ_serv.lower() in host.services:
            dst_ids.append(host.minion_id)
        else:
            #print(f'{targ_serv} != {host.services} *NOT FOUND!*')
            continue
    return dst_ids

def get_direction(env,source,targ_serv,service): #gets hosts by appliance type. Ex. All brokers, then scans via salt
    get_targets = getType(env,targ_serv) #list of minionids by appliance type, Ex. All broker minionids
    temp_ports = list(Appliance_Sockets[targ_serv].values())
    ports = temp_ports[0]
    report_dict = {}
    report_list_passed = []
    for dst_minionid in get_targets:
        pass_result = direction_scan(env,targ_serv, source, dst_minionid, ports,service)
        report_list_passed = pass_result
        source_host = magical.minion_id_to_hostname(env.hosts, source)
        target_host = magical.minion_id_to_hostname(env.hosts, dst_minionid)
        dict_key = (f'{source_host} ({service}) \u2192 {target_host} ({targ_serv})')
        report_dict[dict_key] = report_list_passed
    return report_dict

def direction_get_targets(env,service): #provides list of target appliances based on provided service name
    direction_constants = Direction
    direction_constants_dict = []
    direction_json = json.dumps(direction_constants)
    direction_constants_dict = json.loads(direction_json)
    targets = []
    for serv, serv_targs in direction_constants_dict.items():
        if service == serv:
            targets = serv_targs
    return targets
        
def directional(env): #function loops through hosts in environment and scans
    direction_dict = {}
    report_dict = {}
    for host in env.hosts:
        #print(f'{host.services}')
        key = str(host.services)[2:-2]
        direction_dict.setdefault(key, [])
        direction_dict[key].append(str(host.minion_id))
    for service, saltid in direction_dict.items(): #environment hosts
        for minion_id in saltid: #list of minion ids
            source = str(minion_id)
            use_targets = direction_get_targets(env,service) 
            for targ_serv in use_targets:
                try:
                    complete_report_dict = get_direction(env,source,targ_serv,service)
                    report_dict.update(complete_report_dict)
                except KeyError: #exception handling for appliance types that the source appliances needs to poke but may not exist in all customer environments
                    continue 
    return report_dict
            
def fw_check(env): # main function, gathers input, calls scan()
    targets = {}
    report_dict = defaultdict(list)
    question = input("Do you have text file? (y/n)")
    if question == "y":
        filepath = input("Enter path to text file: ")
        cleanpath = os.path.abspath(filepath)
        assert os.path.exists(cleanpath), "Please check spelling. I did not find the file at, "+str(cleanpath) #verifies path exists
        f = open(cleanpath, "r+")
        for line in f.readlines():
            targ_ipv4, service_name = line.split(',',1)
            targets[targ_ipv4] = service_name
    else:
        targ_ip = input("Enter target appliance IPv4 address: ")
        target_data = sub_menu()     
        service_num = input("\nWhat is the appliance type? Type corresponding number: ")
        use_service_num = int(service_num)
        choice = list(target_data.keys())[use_service_num]
        targets[targ_ip] = choice 
    appsock = Appliance_Sockets
    appsock_dict = []
    used_ports = []
    appsock_json = json.dumps(appsock)
    appsock_dict = json.loads(appsock_json)
    for app_ipv4, service_name in targets.items(): #dictionary hold target
        name = service_name.strip()
        for app_type, app_ports in appsock_dict.items():
            if name == app_type:
                for ignore_key, use_ports in app_ports.items():
                    used_ports = use_ports
                listener_ports = [i for n, i in enumerate(used_ports) if i not in used_ports[:n]]
                if name.lower() == "search":
                    name = "H&W Beta"
                elif name.lower() == "inclogdecoder":
                    name = "Log Hybrid Retention"
                click.echo(color.blue(f'\nTesting: {app_ipv4} as a {name}'))
                alive_res = [] 
                caution_res = []
                dead_res = []
                for tport in listener_ports:
                    test_port = str(tport)
                    alive, caution, dead = ip_scan(env,app_ipv4,test_port,failcheck=False) #get variables from scan function
                    if alive.isdigit(): #check that returned variable is int and not string
                        alive_res.append(alive)
                    if caution.isdigit(): #check that returned variable is int and not string
                        caution_res.append(caution)
                    if dead.isdigit(): #check that returned variable is int and not string
                        dead_res.append(dead)
                if not alive_res:
                    #fail_test.clear()
                    click.echo(color.yellow("All ports returned closed, switching scan! Hold please."))
                    alive_res.clear()
                    caution_res.clear()
                    dead_res.clear()
                    for tport in listener_ports:
                        test_port = str(tport)
                        alive, caution, dead = ip_scan(env,app_ipv4,test_port,failcheck=True)
                        if alive.isdigit():
                            alive_res.append(alive)
                        if caution.isdigit():
                            caution_res.append(caution)
                        if dead.isdigit():
                            dead_res.append(dead)
                for val in sorted(alive_res):
                    click.echo(color.green(f"{val}: OPEN!"))
                    data = (f"{val} \u2714")
                    report_dict[app_ipv4].append(data)
                for val in sorted(caution_res):
                    click.echo(color.yellow(f"{val}: CONNECTION REFUSED!"))
                    data = (f"{val} \u26A0")
                    report_dict[app_ipv4].append(data)
                for val in sorted(dead_res):
                    click.echo(color.red(f"{val}: CLOSED!"))
                    data = (f"{val} \u274C")
                    report_dict[app_ipv4].append(data)
            else:
                continue  
    return report_dict

def report(env, use_report_dict): #reporting function, will populate html document in root directory of toolkit
    dict_length = len(use_report_dict.keys())
    template_path = f'{pathlib.Path(__file__).parent.absolute()}\\'
    report = Report(template_path, 'portcheck_template.htm')
    report.to_file('portcheck.htm', data = use_report_dict, host_data = env.hosts, table_amnt = dict_length ,overwrite=True)

@click.command(help='Port checker. Test against existing enviroment or use in conjunction with "listener" command')
@click.option('--report-file', 'output_format', flag_value='report_file', help='Outputs results to a file using a report')
@click.pass_obj
def portcheck(env,output_format): #main function
    menu()
    use_report_dict = {}
    option = input("Enter menu option (1 or 2): ")
    start = timeit.default_timer() # <-- used in development/debugging. Can be removed safely
    if "1" in option: #ip scan, no directionality, 1:1 or provide text file path as input
        report_dict = fw_check(env)
    elif "2" in option: #directionality scan, entire environment
        report_dict = directional(env)
    else:
        click.echo("Check your input. Please enter 1 or 2.") #input verification
    for key, value in report_dict.items():
        ports = []
        for port in value:
            port.encode('utf8')
            ports.append(port)
        use_report_dict[key] = ports
    #print(use_report_dict.items()) <-- uncomment to debug report data
    if output_format == "report_file":
        report(env,use_report_dict)
    stop = timeit.default_timer()
    print('Runtime: ', stop - start) # <-- used in development/debugging. Can be removed safely